/* copymod - toggle copy-prevent flags.
 * Public domain by Russell Marks, 2003.
 */

/* all the `#ifdef linux' stuff was for basic debugging on my Linux box;
 * I don't have a DC serial lead so I try to test as much as possible
 * before finally writing a copy for the DC. :-)
 */

#ifdef linux
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <curses.h>
#else
#include <kos.h>
#include <conio/conio.h>
#endif

#define DIR_BLOCK_BASE	241
#define BLOCKS_IN_DIR	13
#define MAX_ENTRIES	45	/* max we can handle displaying */

static unsigned char dirdata[512*BLOCKS_IN_DIR];
static unsigned char olddata[512*BLOCKS_IN_DIR];	/* unmodified copy */
static unsigned char *entryptrs[MAX_ENTRIES];
int num_entryptrs;


#ifdef linux

#define CONT_B			(1<<1)
#define CONT_A			(1<<2)
#define CONT_START		(1<<3)
#define CONT_DPAD_UP		(1<<4)
#define CONT_DPAD_DOWN		(1<<5)
#define CONT_DPAD_LEFT		(1<<6)
#define CONT_DPAD_RIGHT		(1<<7)
#define CONT_Y			(1<<9)
#define CONT_X			(1<<10)

#define conio_init(x,y)
#define conio_shutdown()
#define conio_printf printw

typedef void maple_device_t;


typedef struct { unsigned int buttons; } cont_state_t;

void pvr_init_defaults(void)
{
initscr();
cbreak(); noecho();
nodelay(stdscr,TRUE);
}

maple_device_t *maple_enum_dev() { return NULL; }

int vmu_block_read(maple_device_t *dev,unsigned short num,unsigned char *buf)
{
static FILE *in=NULL;

if(!in)
  in=fopen("fakevmudir.dat","rb");

if(num==DIR_BLOCK_BASE) rewind(in);
fread(buf,1,512,in);
return(0);
}

int vmu_block_write(maple_device_t *dev,unsigned short num,unsigned char *buf)
{
static FILE *out=NULL;

if(!out)
  out=fopen("fakevmudir.new","wb");

fwrite(buf,1,512,out);
fflush(out);
return(0);
}

void *maple_dev_status(maple_device_t *dev)
{
static cont_state_t state;
int key;

refresh();
key=getch();

state.buttons=0;
switch(key)
  {
  case 'q': state.buttons|=CONT_DPAD_UP; break;
  case 'a': state.buttons|=CONT_DPAD_DOWN; break;
  case 'o': state.buttons|=CONT_DPAD_LEFT; break;
  case 'p': state.buttons|=CONT_DPAD_RIGHT; break;
  case 'y': state.buttons|=CONT_Y; break;
  case 'b': state.buttons|=CONT_B; break;
  case 32: state.buttons|=CONT_A; break;
  }

return(&state);
}

void conio_clear(void)
{
int f;
clear();

for(f=0;f<18;f++)
  mvaddch(f,48,'|');
mvaddstr(18,0,"------------------------------------------------+");
move(0,0);
}

void conio_gotoxy(int x,int y)
{
move(y,x);
}

void thd_sleep(int ms)
{
usleep(ms*1000);
}

#endif


#define MAIN_BUTTONS	(CONT_B|CONT_A|CONT_START| \
			 CONT_DPAD_UP|CONT_DPAD_DOWN| \
			 CONT_DPAD_LEFT|CONT_DPAD_RIGHT| \
			 CONT_Y|CONT_X)

void waitfornobuttons(maple_device_t *paddev)
{
cont_state_t *padstate;

do
  {
  thd_sleep(10);
  padstate=maple_dev_status(paddev);
  }
while(!padstate || (padstate->buttons&MAIN_BUTTONS)!=0);
}


/* returns nz if we want to write them */
int askwritechanges(maple_device_t *paddev)
{
cont_state_t *padstate;

conio_clear();
conio_gotoxy(0,0);

/* check for changes */
if(memcmp(olddata,dirdata,sizeof(olddata))==0)
  {
  conio_printf("\n\nNo changes made, no save necessary.\n");
  return(0);
  }

waitfornobuttons(paddev);

conio_printf("\n\nDo you want to save changes to VMU?\n\n");
conio_printf("Press A for yes, or B for no.\n");

while(1)
  {
  thd_sleep(10);
  padstate=maple_dev_status(paddev);
  
  if(padstate && (padstate->buttons&CONT_B))
    return(0);
  if(padstate && (padstate->buttons&CONT_A))
    return(1);
  }
}


#define GOTOPOS(pos,xadd) conio_gotoxy(((pos)%3)*16+(xadd),(pos)/3+3)

void mainloop(maple_device_t *paddev)
{
unsigned char tmpstr[13];
unsigned char *ptr;
int f,g;
int quit=0;
cont_state_t *padstate;
int pos,oldpos;

conio_clear();
conio_gotoxy(0,0);
conio_printf("    Use d-pad & A button to toggle CP flags.\n");
conio_printf(" (\"*\" below = flag is set.)  Press Y when done.\n");

if(!num_entryptrs)
  {
  conio_printf("\n...actually, don't bother - no files on VMU!\n");
  thd_sleep(2000);
  return;
  }
else
  {
  for(f=0;f<num_entryptrs;f++)
    {
    GOTOPOS(f,0);
    ptr=entryptrs[f];
    memcpy(tmpstr,ptr+4,12);
    tmpstr[12]=0;
    for(g=0;g<12;g++)
      if(tmpstr[g]<32 || tmpstr[g]>126)
        tmpstr[g]=32;
    conio_printf(" %c%-12s",ptr[1]?'*':' ',tmpstr);
    }
  }

pos=0;
oldpos=-1;
while(!quit)
  {
  if(pos!=oldpos)
    {
    if(oldpos>=0)
      {
      GOTOPOS(oldpos,0);
      conio_printf(" ");
      }
    
    GOTOPOS(pos,0);
    conio_printf(">");
    GOTOPOS(pos,0);
    oldpos=pos;
    }

  ptr=entryptrs[pos];
  
  padstate=maple_dev_status(paddev);
  if(!padstate)
    {
    thd_sleep(10);
    continue;
    }
  
  if(padstate->buttons&CONT_Y)
    quit=1;
  
  if(padstate->buttons&CONT_A)
    {
    ptr[1]=(ptr[1]?0:0xff);
    GOTOPOS(pos,1);
    conio_printf(ptr[1]?"*":" ");
    GOTOPOS(pos,0);
    }
  
  if(padstate->buttons&CONT_DPAD_LEFT)
    pos--;
  if(padstate->buttons&CONT_DPAD_RIGHT)
    pos++;
  if(padstate->buttons&CONT_DPAD_UP)
    pos-=3;
  if(padstate->buttons&CONT_DPAD_DOWN)
    pos+=3;

  if(pos<0) pos=0;
  if(pos>num_entryptrs-1) pos=num_entryptrs-1;

  if(padstate->buttons)
    waitfornobuttons(paddev);
  else
    thd_sleep(10);
  }
}


void get_entryptrs(void)
{
int f;

num_entryptrs=0;
for(f=0;f<BLOCKS_IN_DIR*512/32;f++)
  {
  if(num_entryptrs>=MAX_ENTRIES) break;
  
  if(dirdata[f*32]==0x33 || dirdata[f*32]==0xcc)
    entryptrs[num_entryptrs++]=dirdata+f*32;
  }
}


void waitexit(void)
{
conio_printf("\n\n(program exited, rebooting in 5 seconds...)\n");

#ifdef linux
refresh();
sleep(5);
nodelay(stdscr,FALSE);
echo(); nocbreak();
endwin();
exit(0);

#else

thd_sleep(5000);

/* ok, actually go for the DC menu. :-) */
arch_menu();

/* shouldn't return, but in case it does... */
while(1) ;
#endif
}


void checkforpad(maple_device_t *dev)
{
#ifndef linux
if(!dev || !(dev->info.functions & MAPLE_FUNC_CONTROLLER))
  {
  conio_printf("No joypad in port A, giving up!\n");
  waitexit();
  }
#endif
}


void checkforvmu(maple_device_t *dev)
{
#ifndef linux
if(!dev || !(dev->info.functions & MAPLE_FUNC_MEMCARD))
  {
  conio_printf("No VMU in port A slot 1, giving up!\n");
  waitexit();
  }
#endif
}


void readvmudir(maple_device_t *dev,unsigned char *data)
{
int f;

for(f=0;f<BLOCKS_IN_DIR;f++)
  if(vmu_block_read(dev,DIR_BLOCK_BASE+f,data+f*512)<0)
    {
    conio_printf("Failed to read block, giving up!\n");
    waitexit();
    }
}


void writevmudir(maple_device_t *dev,unsigned char *data)
{
int f;

for(f=0;f<BLOCKS_IN_DIR;f++)
  if(vmu_block_write(dev,DIR_BLOCK_BASE+f,data+f*512)<0)
    {
    conio_printf("Failed to write block, giving up!\n");
    waitexit();
    }
}


void checkvmuunchanged(maple_device_t *dev)
{
static unsigned char checkdata[512*BLOCKS_IN_DIR];

readvmudir(dev,checkdata);

if(memcmp(olddata,checkdata,sizeof(olddata))!=0)
  {
  conio_printf("VMU was changed, giving up!\n");
  waitexit();
  }
}


int main(void)
{
maple_device_t *paddev=NULL;
maple_device_t *vmudev=NULL;

#ifndef linux
/* KOS doesn't use 50Hz automatically, so we check for it ourselves. */
if(flashrom_get_region()==FLASHROM_REGION_EUROPE &&
   vid_check_cable()!=CT_VGA)
  vid_init(DM_640x480_PAL_IL,PM_RGB565);	/* 50Hz */
else
  vid_init(DM_640x480,PM_RGB565);
#endif

pvr_init_defaults();

conio_init(CONIO_TTY_PVR,CONIO_INPUT_LINE);

paddev=maple_enum_dev(0,0);
checkforpad(paddev);
vmudev=maple_enum_dev(0,1);
checkforvmu(vmudev);
conio_printf("Reading VMU, please wait...\n");
readvmudir(vmudev,dirdata);
memcpy(olddata,dirdata,sizeof(olddata));

get_entryptrs();

mainloop(paddev);

if(!askwritechanges(paddev))
  conio_printf("\nOk, not writing changes to VMU.\n");
else
  {
  /* only if we need to (and ask to!) write changes... */
  checkforvmu(vmudev);
  checkvmuunchanged(vmudev);
  conio_printf("\nWriting VMU, please wait...\n");
  writevmudir(vmudev,dirdata);
  conio_printf("VMU written ok.\n");
  }

conio_printf("\nFinished, program exiting.\n");
waitexit();

return 0;
}
